#include "CBasisInternalHeapMng.h"
#include "CBasisImmixBlockPool.h"
#include "CBasisMarkRegionConfig.h"
#include "CBasisMemStats.h"
#include "CBasisTiming.h"
#include <memory.h>
#include <stdlib.h>
#include <stdio.h>

// Immix block line size
#define IBP_LINE_SIZE     MRC_LINE_SIZE

// Immix line bitmap size in bits
#define IBP_LINE_BITMAP_SIZE_BITS (IHM_BLOCK_SIZE / IBP_LINE_SIZE)

// Immix line bitmap size in bytes
#define IBP_LINE_BITMAP_SIZE (IBP_LINE_BITMAP_SIZE_BITS / 8)

// Immix blocks to allocate at a time
#define IBP_BLOCKS_AT_A_TIME MRC_BLOCKS_AT_A_TIME

#define IBP_BLOCK_PAYLOAD_SIZE (IHM_BLOCK_SIZE - IBP_LINE_SIZE)

// The clear mask when performing a full heap collection
#define IBP_CLEAR_MASK_FULL_COLLECTION         ((unsigned int)~0x1)

// The clear mask when performing only a partial heap collection
#define IBP_CLEAR_MASK_PARTIAL_COLLECTION      ((unsigned int)~0x0)

// The global lists of allocated blocks,
// classified as unavailable, recycled or free
void *ibpUnavailableBlockPool = NULL;
void *ibpRecycledBlockPool = NULL;
void *ibpFreeBlockPool = NULL;

// Bump pointer allocation
char *currentPtr = NULL;
char *endPtr = NULL;

// Memory of the previously initiated (i.e. currently executing) collection
unsigned int ibpInitiatedClearMask = IBP_CLEAR_MASK_FULL_COLLECTION;
int ibpCompletedLastCollection = 1;

void IbpInitImmixBlock(void *block) {
    // Clear the line bitmap and the block list pointer
    memset(block, 0, IBP_LINE_BITMAP_SIZE + sizeof(void *));
}

void IbpClearMarksImmixBlock(void *block) {
    unsigned int *lineBitmap = (unsigned int *)block;
    //printf("Clearing block at %d\n", (int)block);
    // Reset 256 bits (32768 / 128 lines), taking up 256 / 32 = 8 32-bit words
    *(lineBitmap + 0) = 0;
    *(lineBitmap + 1) = 0;
    *(lineBitmap + 2) = 0;
    *(lineBitmap + 3) = 0;
    *(lineBitmap + 4) = 0;
    *(lineBitmap + 5) = 0;
    *(lineBitmap + 6) = 0;
    *(lineBitmap + 7) = 0;
}

// Returns 1 if block is in use (i.e. at least one bit is marked), 0 otherwise
int IbpHasMarksImmixBlock(void *block) {
    unsigned int *lineBitmap = (unsigned int *)block;
    unsigned int commonBitmap =
	*(lineBitmap + 0) |
	*(lineBitmap + 1) |
	*(lineBitmap + 2) |
	*(lineBitmap + 3) |
	*(lineBitmap + 4) |
	*(lineBitmap + 5) |
	*(lineBitmap + 6) |
	*(lineBitmap + 7);

    // This makes the returned value compatible with the compiler's bool true
    return (int)(commonBitmap != 0);
}

// Returns 1 if block is full (i.e. all bits are marked), 0 otherwise
int IbpIsMarkedFullImmixBlock(void *block) {
    unsigned int *lineBitmap = (unsigned int *)block;
    unsigned int commonBitmap =
	*(lineBitmap + 0) &
	*(lineBitmap + 1) &
	*(lineBitmap + 2) &
	*(lineBitmap + 3) &
	*(lineBitmap + 4) &
	*(lineBitmap + 5) &
	*(lineBitmap + 6) &
	*(lineBitmap + 7);

    // We OR with 1 because the first line in the bitmap contains header
    // information and is therefore never used, and because we can never
    // be sure to use the first line marked in the first bit of the
    // following 7 words in the line mark bitmap, due to conservative marking.
    // This makes the returned value compatible with the compiler's bool true
    // The constant is 32^2 - 1
    return (int)((commonBitmap | 1) == (int)(-1));
}

// Show the line marks of a given block
void IbpShowImmixBlockLineMarks(void *block) {
    int word = 0;
    int bitNumber = 0;
    unsigned int bitMask  = (((unsigned int)0x1) << bitNumber);
    unsigned int *lineMarkWord = (unsigned int *)block;
    while((char *)lineMarkWord < (((char *)block) + IBP_LINE_BITMAP_SIZE)) {
        while(bitNumber < 32) {
            bitMask = (unsigned int)(0x1 << bitNumber);
	    if((*lineMarkWord & bitMask) == 0) {
                printf(".");
	    } else {
                printf("1");
	    }

            // Advance to next bit
	    bitNumber++;
	}

	// Advance to next 32-bit word (4 bytes) and start at first two bits
        bitMask = 0x1;
        bitNumber = 0;
	lineMarkWord++;
	word++;
	printf("|\n");
    }
}

void *IbpGetImmixBlockAreaStart(void *block) {
    return (void *)(((char *)block) + IBP_LINE_SIZE);
}

void *IbpGetImmixBlockAreaEnd(void *block) {
    return (void *)(((char *)block) + IHM_BLOCK_SIZE);
}

void *IbpGetImmixBlockListPtr(void *block) {
    return *((void **)(((char *)block) + IBP_LINE_BITMAP_SIZE));
}

void IbpSetImmixBlockListPtr(void *block, void *ptr) {
    *((void **)(((char *)block) + IBP_LINE_BITMAP_SIZE)) = ptr;
}

void *IbpInsertImmixBlock(void *block, void *blockPool) {
    // This is like a Standard ML list cons, where blockPool is a pointer to
    // the first block in the list and block is the new block to be cons'ed
    IbpSetImmixBlockListPtr(block, blockPool);
    return block;
}

// Allocate initialize and put some blocks on the free list
void *IbpAllocFreshBlocks() {
    //puts("Allocating fresh blocks");
    char *blocks = IhmAllocAlignedBlocks(IBP_BLOCKS_AT_A_TIME);
    void *block = NULL;
    int cnt;
    for(cnt = 0; cnt < IBP_BLOCKS_AT_A_TIME; cnt++) {
	// Put the blocks onto the free list, such that they are
        // listed avaiable in consecutive allocation order
	block = (void *)(blocks + (IBP_BLOCKS_AT_A_TIME - cnt - 1) * IHM_BLOCK_SIZE);
	IbpInitImmixBlock(block);
	//printf("IbpAllocFreshBloks: IbpInsertImmixBlock(%d, %d)\n", (int)block, (int)ibpFreeBlockPool);
	ibpFreeBlockPool = IbpInsertImmixBlock(block, ibpFreeBlockPool);
    }

    // Check that things went as we expect
    if(blocks != ibpFreeBlockPool) {
        puts("Expecting the address of the allocated blocks to be the same as the final free list address");
    }

    return ibpFreeBlockPool;
}

void *IbpGetFreeBlock() {
    void *result;

    //puts("IbpGetFreeBlock");
    if(!ibpFreeBlockPool) {
        // There are no free blocks in the free pool, allocate some
	IbpAllocFreshBlocks();
    }

    // There are blocks in the free pool, get one
    result = ibpFreeBlockPool;
    //printf("Got free block at address %d\n", (int)result);

    // Remove block from free pool
    ibpFreeBlockPool = IbpGetImmixBlockListPtr(result);
    // Add block to unvailable pool
    //printf("IbpGetFreeBlock: IbpInsertImmixBlock(%d, %d)\n", (int)result, (int)ibpUnavailableBlockPool);
    ibpUnavailableBlockPool = IbpInsertImmixBlock(result, ibpUnavailableBlockPool);

    //puts("Inserted block");
    // Initialize bump allocation pointers
    currentPtr = IbpGetImmixBlockAreaStart(result);
    endPtr = IbpGetImmixBlockAreaEnd(result);

    // Clear the memory area, for
    // debugging if anything important is overwritten
    //memset(currentPtr, 0, IBP_BLOCK_PAYLOAD_SIZE - IBP_LINE_SIZE);

    //puts("Cleared block");
    

    return result;
}

// Find an allocation span with at least bytes free bytes available
void IbpGetLinesWithFreeBytes(void *block, void *startPtr, int bytes) {
    //printf("IbpGetLinesWithFreeBytes %d\n", (int)bytes);
    if((unsigned int)(startPtr - block) >= IHM_BLOCK_SIZE) {
	puts("Something completely wrong with the parameters for IbpGetLinesWithFreeBytes");
        exit(-1);
    }
    unsigned int word = ((unsigned int)(startPtr - block)) >> (5 + 7);
    // Bit number is the first of two bits to test, due to conservative line marking
    unsigned int bitNumber = ((((unsigned int)startPtr) >> 7) & 0x1f);
    if(bitNumber == 0) {
	bitNumber = 1;
    }
    unsigned int bitsMask  = (((unsigned int)0x1) << bitNumber);
    unsigned int *lineMarkWord = ((unsigned int *)block) + word;
    while((char *)lineMarkWord < (((char *)block) + IBP_LINE_BITMAP_SIZE)) {
        // Continue searching for a bit until the end of the word
        //printf("Searching word %d\n", (int)word);
	while(bitNumber < 32) {
            //printf("Searching bit %d\n", (int)bitNumber);
	    if((*lineMarkWord & bitsMask) == 0) {
                // A bit is free, which
		// is a potential start address for allocation
		currentPtr =
		    ((char *)block) +
		    (word * 32 + bitNumber) * IBP_LINE_SIZE;
		if(currentPtr < (char *)startPtr) {
		    puts("currentPtr < startPtr");
                    exit(-1);
		}
		if((currentPtr < (char *)block) ||
		   currentPtr > ((char *)block + IHM_BLOCK_SIZE)) {
		    puts("currentPtr out of block");
                    exit(-1);
		}

		// Now continue to find an end address
		// Notice: This is usable as a quick and dirty test:
	        // endPtr = currentPtr + IBP_LINE_SIZE;
                
		// Advance to next bit
	        bitNumber++;
		bitsMask = bitsMask << 1;

                // Continue searching for a marked bit until the end of the block
		while((char *)lineMarkWord < (((char *)block) + IBP_LINE_BITMAP_SIZE)) {
                    // Continue searching for a bit until the end of the word
		    while(bitNumber < 32) {
                        //printf("Searching bit %d\n", (int)bitNumber);
            	        if((*lineMarkWord & bitsMask) != 0) {
			    // We encountered a dirty bit, compute the span
			    // we have found and check if it is good enugh
                            endPtr =
	                        ((char *)block) +
			        (word * 32 + bitNumber) * IBP_LINE_SIZE;

			    // Found an allocation span
                            // Check if the area found is big enough
		            if((unsigned int)(endPtr - currentPtr) < bytes) {
		                // We found an area which is not big enough,
		                // continue searching recursively
                                IbpGetFreeAllocationSpan(bytes);
		            }

                            return;
			}

                        // Advance to next bit
	                bitNumber++;
                        bitsMask = bitsMask << 1;
		    }

                    // Advance to next 32-bit word (4 bytes) and start at second bit
                    bitsMask = 0x1;
                    bitNumber = 0;
	            lineMarkWord++;
                    word++;
		}

		// We have now searched the entire block for a dirty end bit,
		// but found none, so we can allocate into the rest of
                // the block. endPtr is thus the end of the block
		endPtr = ((char *)block) + IHM_BLOCK_SIZE;

		// Check if the area found is big enough
		if((unsigned int)(endPtr - currentPtr) < bytes) {
		    // We found an area which is not big enough,
		    // continue searching recursively
                    IbpGetFreeAllocationSpan(bytes);
		}

		// Clear the memory area, for
		// debugging if anything important is overwritten
		//memset(currentPtr, 0, IBP_LINE_SIZE);

		//printf("Found bit %d in block %d, currentPtr %d, endPtr %d\n", (int)bitNumber, (int)block, (int)currentPtr, (int)endPtr);

		return;
	    }

            // Advance to next bit
	    bitNumber++;
            bitsMask = bitsMask << 1;
	}

	// Advance to next 32-bit word (4 bytes) and start at second bit
        bitsMask = 0x2;
        bitNumber = 1;
	lineMarkWord++;
        word++;
    }

    // Exhausted the search in this block, get a new block and search there
    IbpGetBlockWithFreeBytes(bytes);
}

// Find a block and an allocation span with at least bytes free bytes available
void IbpGetBlockWithFreeBytes(int bytes) {
    void *recycledBlock;

    //printf("IbpGetBlockWithFreeBytes %d\n", (int)bytes);
    if(ibpRecycledBlockPool) {
        //puts("ibpRecycledBlockPool is a pointer (not NULL)");
	// There are blocks in the recycled pool, get one
	recycledBlock = ibpRecycledBlockPool;
	//printf("Got recycled block at address %d\n", (int)recycledBlock);

	// Remove block from recycled pool
	ibpRecycledBlockPool = IbpGetImmixBlockListPtr(recycledBlock);

	// Add block to unvailable pool
	//printf("IbpGetBlockWithFreeBytes: IbpInsertImmixBlock(%d, %d)\n", (int)recycledBlock, (int)ibpUnavailableBlockPool);
	ibpUnavailableBlockPool = IbpInsertImmixBlock(recycledBlock, ibpUnavailableBlockPool);

	// Initialize bump allocation pointers. When giving it the
	// start of the block as initial search pointer, it will
        // skip the initial header line, due to conservative line marking
        IbpGetLinesWithFreeBytes(recycledBlock, IbpGetImmixBlockAreaStart(recycledBlock), bytes);
    } else {
        //puts("ibpRecycledBlockPool is NULL");
	// There are no more recycled blocks, get a free block
        IbpGetFreeBlock();
    }
    //puts("Got block with free bytes");
}

// currentPtr is about to pass endPtr, find a new allocation span
void IbpGetFreeAllocationSpan(int bytes) {
    //printf("IbpGetFreeAllocationSpan %d\n", (int)bytes);
    if((((unsigned int)endPtr) & (IHM_BLOCK_SIZE - 1)) == 0) {
        //printf("endPtr is aligned with block, at %d\n", (int)endPtr);
	// endPtr is aligned with a block, so this is the end of a block.
	// Find a new block, either recycled, if there are any, or a free block
        IbpGetBlockWithFreeBytes(bytes);
    } else {
	//printf("endPtr is not aligned with block, at %d\n", (int)endPtr);
	// endPtr is not aligned with a block, find next free hole of
	// lines within the current block
	IbpGetLinesWithFreeBytes((unsigned int *)(((unsigned int)endPtr) & (~(IHM_BLOCK_SIZE - 1))), endPtr, bytes);
    }
    //puts("Got free allocation span");
}

// The final managed allocation function
void *IbpAlloc(int bytes) {
    void *result;

    //printf("IbpAlloc(%d)\n", (int)bytes);
    // We don't want to allocate zero bytes, ever
    // FIXME: Change the translation instead to avoid this
    if(bytes == 0) {
        bytes = 4;
    }

    // FIXME: Handle allocation of large objects too...
    if((unsigned int)(endPtr - currentPtr) >= bytes) {
        // We can directly allocate the desired amount of memmory
	result = currentPtr;
	currentPtr += bytes;
	return result;
    } else {
        //printf("We need more memory because %d - %d = %d < %d\n", (int)endPtr, (int)currentPtr, (int)(endPtr - currentPtr), (int)bytes);
	// We need more memory
	// FIXME: Check if it is time to trigger a collection...
	// Acquire memory
        IbpGetFreeAllocationSpan(bytes);
	//IbpGetFreeBlock();

	if((unsigned int)(endPtr - currentPtr) < bytes) {
	    printf("Unable to allocate %d bytes of memory. Exiting to avoid crashes\n", (int)bytes);
            exit(-1);
	}

        // Bump allocate the block of memory
	result = currentPtr;
	currentPtr += bytes;
        return result;
    }
}

// Clear line mark bitmaps in a linked list of mark-region blocks
void IbpClearLineMarkBitmapsInList(void *blockList) {
    char *blockPtr = (char *)blockList;

    while(blockPtr != NULL) {
        //printf("IbpClearMarksImmixBlock(%d)\n", (int)blockPtr);
	IbpClearMarksImmixBlock(blockPtr);
        blockPtr = IbpGetImmixBlockListPtr(blockPtr);
    }
}

// Clear all the line mark bitmaps of blocks which are in use
void IbpClearLineMarkBitmapsInUse(unsigned int clearMask) {
    //printf("IbpClearLineMarkBitmapsInList(%d)\n", (int)ibpUnavailableBlockPool);
    BtiStartPauseTimer();

#ifdef MRC_ENABLE_INCREMENTAL
    // Start the timer for incremental garbage collection termination
    //printf("\n(%d)S", (int)clearMask);
    BtiStartQuickAsyncTimerMicro(MRC_INCREMENTAL_TIMING_MICRO_SECONDS);
    if(ibpCompletedLastCollection) {
    ibpCompletedLastCollection = 0;
#endif
	
#ifdef MRC_ENABLE_IN_PLACE_GENERATIONAL
    if(clearMask == IBP_CLEAR_MASK_FULL_COLLECTION) {
        // Line-mark bitmaps should only be cleared on full-heap allocations
	IbpClearLineMarkBitmapsInList(ibpUnavailableBlockPool);
    }
#else
    // When in-place generational GC is enabled, collections are
    // always full-heap collections, so all mark bits are always cleared
    IbpClearLineMarkBitmapsInList(ibpUnavailableBlockPool);
#endif
#ifdef MRC_ENABLE_INCREMENTAL
    }
#endif // MRC_ENABLE_INCREMENTAL
    //puts("Cleared");
    // We should always make sure that there are no recycled blocks,
    // when initiating a full GC, which is the only time where
    // the line mark bitmaps are fully cleared with this function
    //IbpClearLineMarkBitmapsInList(ibpRecycledBlockPool);
}

// The function that garbage collection is all about
void IbpReclaimGarbage() {
    char *newUnavailableBlocks = (char *)NULL;
    char *blockPtr = (char *)ibpUnavailableBlockPool;
    char *nextBlockPtr = (char *)NULL;

#ifdef MRC_ENABLE_INCREMENTAL
    if(!BtiIsQuickAsyncTimerFired()) {
        // The incremental tracing has completed, so it is safe to reclaim garbage
#endif // MRC_ENABLE_INCREMENTAL

    //puts("Reclaiming garbage");
    if(ibpInitiatedClearMask == IBP_CLEAR_MASK_FULL_COLLECTION) {
	// We only reset the conservative liveness estimate when
        // initiating full-heap collections
	BmsRegisterResetConservativeLive();
    }

    while(blockPtr != NULL) {
        //printf("Line mark bitmap of block at %d\n", (int)blockPtr);
        //IbpShowImmixBlockLineMarks(blockPtr);

        // Remember pointer to next block in list, before it is overwritten
	nextBlockPtr = IbpGetImmixBlockListPtr(blockPtr);

	if(IbpHasMarksImmixBlock(blockPtr)) {
	    // The block is still in use, check if it is recyclable
	    if(IbpIsMarkedFullImmixBlock(blockPtr)) {
                // The block is full, retain it in the unavailable list
	        //printf("Keeping block at %d\n", (int)blockPtr);
	        newUnavailableBlocks =
	            IbpInsertImmixBlock(blockPtr, newUnavailableBlocks);

	        BmsRegisterConservativeLiveBytes(IBP_BLOCK_PAYLOAD_SIZE);
	    } else {
		// The block is only partially full, recycle it

		// Check if we are already allocating into the recycled block
		if((currentPtr >= (char *)IbpGetImmixBlockAreaStart(blockPtr)) &&
		   (currentPtr <= (char *)IbpGetImmixBlockAreaEnd(blockPtr))) {
		    // We are already allocating into this block, which is
		    // actually perfect. Just continue doing that
		    // and mark it as unavailable, rather than recycled
		    newUnavailableBlocks =
		        IbpInsertImmixBlock(blockPtr, newUnavailableBlocks);
		} else {
		    ibpRecycledBlockPool =
			IbpInsertImmixBlock(blockPtr, ibpRecycledBlockPool);
		}

                // This estimate is very conservative...
		BmsRegisterConservativeLiveBytes(IBP_BLOCK_PAYLOAD_SIZE);
	    }
	} else {
            // Block is unused, reclaim it
            //printf("Reclaiming block at %d\n", (int)blockPtr);
	    ibpFreeBlockPool = IbpInsertImmixBlock(blockPtr, ibpFreeBlockPool);
	    // Clear the memory area, for
	    // debugging if anything important is overwritten
	    //memset(IbpGetImmixBlockAreaStart(blockPtr), 0,
	    //      IBP_BLOCK_PAYLOAD_SIZE - IBP_LINE_SIZE);

	    // Check that we were not allocating into the now reclaimed block,
	    // since it would not be valid to continue with that when it
            // is marked as free
	    if((currentPtr >= (char *)IbpGetImmixBlockAreaStart(blockPtr)) &&
	       (currentPtr <= (char *)IbpGetImmixBlockAreaEnd(blockPtr))) {
		// We were allocating into this block, reset bump allocation
		currentPtr = NULL;
                endPtr = NULL;
	    }
	    BmsRegisterReclaimedBlock();
	}

        // Iterate to next block in list, by using the saved next pointer
	blockPtr = nextBlockPtr;
    }

    // An updated list of used blocks has now been constructed
    ibpUnavailableBlockPool = newUnavailableBlocks;

    // Make sure that bump allocation pointers are properly re-initialized
    if(currentPtr == NULL) {
	// Get a free block to allocate into
	// FIXME: Try to get a recycled block instead...
	IbpGetFreeBlock();
    }

    //puts("Reclaimed garbage");
    BmsRegisterCollectionComplete();
    if(ibpInitiatedClearMask == IBP_CLEAR_MASK_FULL_COLLECTION) {
	BmsRegisterFullCollectionComplete();
    } else {
	BmsRegisterPartialCollectionComplete();
    }
#ifdef MRC_ENABLE_INCREMENTAL
        //printf("C");
        ibpCompletedLastCollection = 1;
    } // End of incremental GC completed
    //else {
    //    printf("I");
    //}
#endif // MRC_ENABLE_INCREMENTAL
    // We have to measure pause times on incremental and completed collections
    BtiReadPauseTimerMicro();
}

// Returns 1 if we should reclaim garbage, 0 otherwise
int IbpShouldReclaimGarbage() {
    int conservativeLive = BmsGetConservativeLive();
    int prevAllocBytes = BmsGetPrevCollectionAllocBytes();
    int allocBytes = BmsGetAllocBytes();

    // Compute the newly allocated number of bytes since the last full collection
    int newlyAllocBytes = allocBytes - prevAllocBytes;

    // Compute how much we estimate has been allocated in the heap
    int heapEstimate = 0;

    // Make sure that we at least consider one block as being live always
    if(conservativeLive <= 0) {
        conservativeLive = IBP_BLOCK_PAYLOAD_SIZE;
    }

    // Compute how much we estimate has been allocated in the heap
    heapEstimate = conservativeLive + newlyAllocBytes;

#ifdef MRC_NEVER_GARBAGE_COLLECT
    return 0;
#else
    // We want to reclaim garbage, if the heap is estimated to have allocated
    // 3 times as much memory as was conservatively live on previous collection
    int shouldReclaimFullHeap =
	((conservativeLive * 3) < heapEstimate) &&
	(ibpRecycledBlockPool == NULL);

#ifdef MRC_ENABLE_IN_PLACE_GENERATIONAL
#ifdef MRC_IN_PLACE_GENERATIONAL_PARTIAL_WORK_BASED_BYTES
    // Compute the newly allocated number of bytes
    int prevPartialAllocBytes = BmsGetPartialPrevCollectionAllocBytes();
    int newlyPartialAllocBytes = allocBytes - prevPartialAllocBytes;

    // Return true if we should do either a full-heap or a partial collection
    return shouldReclaimFullHeap ||
	// This partial heap collection policy is to do a partial collection,
        // if we have allocated more than previously conservatively live
        (newlyPartialAllocBytes * 3 > conservativeLive * 1);// > MRC_IN_PLACE_GENERATIONAL_PARTIAL_WORK_BASED_BYTES);
#endif // MRC_ENABLE_IN_PLACE_GENERATIONAL
#else
    // Only check for full-heap collections
    return shouldReclaimFullHeap;
#endif // MRC_IN_PLACE_GENERATIONAL_PARTIAL_WORK_BASED_BYTES
#endif // MRC_NEVER_GARBAGE_COLLECT
}

// Return the tracing-function clear-mask for either a full
// or a partial heap collection, depending on what it is time for
unsigned int IbpFullOrPartialClearMask() {
    // Notice: This is ugly cut and paste, but at least the
    // IbpShouldReclaimGarbage _should_ be efficient and not
    // call any functions, since it is called _very_ frequently.
    // This function may be less efficient, as it is called only
    // when actually initiating collections
    int conservativeLive = BmsGetConservativeLive();
    int prevAllocBytes = BmsGetPrevCollectionAllocBytes();
    int allocBytes = BmsGetAllocBytes();

    // Compute the newly allocated number of bytes
    int newlyAllocBytes = allocBytes - prevAllocBytes;

    // Compute how much we estimate has been allocated in the heap
    int heapEstimate = 0;

    // Make sure that we at least consider one block as being live always
    if(conservativeLive <= 0) {
        conservativeLive = IBP_BLOCK_PAYLOAD_SIZE;
    }

    // Compute how much we estimate has been allocated in the heap
    heapEstimate = conservativeLive + newlyAllocBytes;

    //printf("conservativeLive: %d, newlyAllocBytes: %d, heapEstimate: %d\n",
    //       (int)conservativeLive, (int)newlyAllocBytes, (int)heapEstimate);

    // We want to reclaim the full heap, if the heap is estimated to
    // have allocated 3 times as much memory as was conservatively
    // live on previous collection
    int shouldReclaimFullHeap =
	((conservativeLive * 3) < heapEstimate) &&
	(ibpRecycledBlockPool == NULL);

#ifdef MRC_IN_PLACE_GENERATIONAL_PARTIAL_WORK_BASED_BYTES
#ifdef MRC_ENABLE_INCREMENTAL
    if(ibpCompletedLastCollection) {
	//ibpCompletedLastCollection = 0;
#endif // MRC_ENABLE_INCREMENTAL
	if(shouldReclaimFullHeap) {
            ibpInitiatedClearMask = IBP_CLEAR_MASK_FULL_COLLECTION;
        } else {
            ibpInitiatedClearMask = IBP_CLEAR_MASK_PARTIAL_COLLECTION;
	}
#ifdef MRC_ENABLE_INCREMENTAL
    } else {
        ibpInitiatedClearMask = IBP_CLEAR_MASK_PARTIAL_COLLECTION;
    }
#endif // MRC_ENABLE_INCREMENTAL
#else
#ifdef MRC_ENABLE_INCREMENTAL
    // Always do full-heap collections if partial collection is disabled
    if(ibpCompletedLastCollection) {
        //ibpCompletedLastCollection = 0;
#endif // MRC_ENABLE_INCREMENTAL
	ibpInitiatedClearMask = IBP_CLEAR_MASK_FULL_COLLECTION;
#ifdef MRC_ENABLE_INCREMENTAL
    } else {
        ibpInitiatedClearMask = IBP_CLEAR_MASK_PARTIAL_COLLECTION;
    }
#endif // MRC_ENABLE_INCREMENTAL
#endif // MRC_IN_PLACE_GENERATIONAL_PARTIAL_WORK_BASED_BYTES

    return ibpInitiatedClearMask;
}

// Returns 1 if we should continue tracing back upwards the stack, 0 otherwise
int IbpShouldStillTrace() {
    return !BtiIsQuickAsyncTimerFired();
}


// Mark a line bit corresponding to a given address
void IbpLineMarkAddr(void *addr) {
    unsigned int *lineMarkBitmapAddr = (unsigned int *)(((unsigned int)addr) & (~(IHM_BLOCK_SIZE - 1)));
    unsigned int word = ((unsigned int)(((unsigned int)addr) & (IHM_BLOCK_SIZE - 1))) >> (5 + 7);
    unsigned int *lineMarkWordAddr = (unsigned int *)(lineMarkBitmapAddr + word);
    // FIXME: Or shift from max unsigned int towards the right?
    unsigned int bit  = (((unsigned int)0x3) << ((((unsigned int)addr) >> 7) & 0x1f));

    // Mark the line corresponding to addr
    *lineMarkWordAddr = (*lineMarkWordAddr) | bit;
}
