#include "CBasisMemStats.h"
#include <sys/types.h> // For pid_t (returned from getpid)
#include <unistd.h> // For getpid
#include <stdio.h>

// Some simple statistics on memory usage
int bmsAllocCnt = 0;
long long bmsAllocByteCnt = 0;
int bmsReclaimBlockCnt = 0;
int bmsAllocMaxBytes = 0;
int bmsCollectionCnt = 0;
int bmsFullCollectionCnt = 0;
int bmsPrevCollectionAlloc = 0;
int bmsPartialPrevCollectionAlloc = 0;
int bmsConservativeLive = 0;

// This shows the operating system's view on memory usage (on Linux at least)
// Notice: Using getrusage should have worked, but according to a post
//         on a website I have seen, that supposedly does not work on
//         e.g. Linux or Solaris (possibly other systems as well)
void BmsShowSysStats() {
    char buf[100];
    snprintf(buf, 99, "/proc/%u/statm", (unsigned)getpid());
    int sz = getpagesize();
    FILE* pf = fopen(buf, "r");
    if (pf) {
        unsigned size; //       total program size
        unsigned resident;//   resident set size
        unsigned share;//      shared pages
        unsigned text;//       code / executable segment
        unsigned lib;//        library (seems to be zero?)
        unsigned data;//       data/stack
        //unsigned dt;//         dirty pages (unused in Linux 2.6)
        fscanf(pf, "%u %u %u %u %u %u", &size, &resident, &share, &text, &lib, &data);
	printf("Linux's view on memory usage, from %s:\n", buf);
	printf("Used memory                   : %dkb / %dmb / %dgb\n",
	       (int)((size * sz) / 1024),
	       (int)((size * sz) / (1024 * 1024)),
	       (int)((size * sz) / (1024 * 1024 * 1024)));
        printf("Resident set kilo bytes       : %d\n", (int)((resident * sz) / 1024));
	printf("Shared kilo bytes             : %d\n", (int)((share * sz) / 1024));
	printf("Code / executable segment kb  : %d\n", (int)((text * sz) / 1024));
	//printf("Library                       : %d\n", (int)((lib * sz) / 1024));
        printf("Data segment size kilo bytes  : %d\n", (int)((data * sz) / 1024));
    }
    fclose(pf);
}

// Reset statistical information
void BmsResetStats() {
    bmsAllocCnt = 0;
    bmsAllocByteCnt = 0;
    bmsReclaimBlockCnt = 0;
    bmsAllocMaxBytes = 0;
    bmsCollectionCnt = 0;
    bmsFullCollectionCnt = 0;
    bmsPrevCollectionAlloc = 0;
    bmsPartialPrevCollectionAlloc = 0;
    bmsConservativeLive = 0;
}

void BmsRegisterAllocatedBytes(int bytes) {
    bmsAllocCnt++;
    bmsAllocByteCnt += bytes;
    if(bytes > bmsAllocMaxBytes) {
        bmsAllocMaxBytes = bytes;
    }
}

void BmsRegisterReclaimedBlock() {
    bmsReclaimBlockCnt++;
}

void BmsRegisterResetConservativeLive() {
    bmsConservativeLive = 0;
}

void BmsRegisterConservativeLiveBytes(int bytes) {
    bmsConservativeLive += bytes;
}

void BmsRegisterCollectionComplete() {
    bmsCollectionCnt++;
}

void BmsRegisterFullCollectionComplete() {
    bmsFullCollectionCnt++;
    bmsPrevCollectionAlloc = bmsAllocByteCnt;
    bmsPartialPrevCollectionAlloc = bmsAllocByteCnt;
}

void BmsRegisterPartialCollectionComplete() {
    bmsPartialPrevCollectionAlloc = bmsAllocByteCnt;
}

// Functions for querying statistical information
int BmsGetConservativeLive() {
    return bmsConservativeLive;
}

int BmsGetPrevCollectionAllocBytes() {
    return bmsPrevCollectionAlloc;
}

int BmsGetPartialPrevCollectionAllocBytes() {
    return bmsPartialPrevCollectionAlloc;
}

int BmsGetAllocBytes() {
    return bmsAllocByteCnt;
}


// Show gathered statistical information and the system's information
void BmsShowStats() {
    printf("Number of allocations         : %d\n", (int)bmsAllocCnt);
    printf("Allocated memory              : %dkb / %dmb / %dgb\n",
	   (int)(bmsAllocByteCnt / 1024),
	   (int)(bmsAllocByteCnt / (1024 * 1024)),
	   (int)(bmsAllocByteCnt / (1024 * 1024 * 1024)));
    printf("Reclaimed regions             : %d\n", (int)bmsReclaimBlockCnt);
    printf("Largest memory block          : %d bytes\n", (int)bmsAllocMaxBytes);
    if(bmsAllocCnt > 0) {
        printf("Average memory block          : %d bytes\n", (int)(bmsAllocByteCnt / bmsAllocCnt));
    }
    printf("Number of collections         : %d\n", (int)bmsCollectionCnt);
    printf("Number of full collections    : %d\n", (int)bmsFullCollectionCnt);
    
    BmsShowSysStats();
}
