#include "CBasisTiming.h"
#include <stdio.h>
#include <stdlib.h>
#include <signal.h> // Used for (the timer_create family of functions?) and signal handling
#include <unistd.h>
#include <time.h> // Used for the clock_gettime family of functions (and the timer_create family of functions?)
#include <sys/time.h> // Used for gettimeofday

long long btiPauseTimer = 0;

int btiQuickAsyncTimerFired = 0;
timer_t btiQuickAsyncTimerId;
struct sigevent btiQuickAsyncTimerSev;
long long btiSlowIncrTimer = 0;

#define BTI_QUICK_ASYNC_TIMER_SIGNAL  SIGRTMIN

// Some simple timing statistics
long long btiMaxPauseTime = 0;
long long btiTotalPauseTime = 0;
long long btiPauseCnt = 0;

void BtiResetTimeStats() {
    btiMaxPauseTime = 0;
    btiTotalPauseTime = 0;
    btiPauseCnt = 0;
}

void BtiShowTimeStats() {
    printf("Max pause time                : %dus / %dms / %ds\n",
	   (int)btiMaxPauseTime,
	   (int)(btiMaxPauseTime / 1000),
	   (int)(btiMaxPauseTime / 1000000));
    printf("Pauses                        : %d\n", (int)btiPauseCnt);
    if(btiPauseCnt > 0) {
	printf("Avg pause time                : %dus / %dms / %ds\n",
	       (int)(btiTotalPauseTime / btiPauseCnt),
	       (int)((btiTotalPauseTime / btiPauseCnt) / 1000),
	       (int)((btiTotalPauseTime / btiPauseCnt) / 1000000));
    }
}

long long BtiStartTimer() {
    // This works and is tested, but is even slower than gettimeofday
    // I believe that it is actually also less portable (not sure though)
    // However, I blieve that it should be more precise
    // (it is also nanoseconds, not microseconds)
    /*
    struct timespec timer;
    // CLOCK_REALTIME could also be used
    if(clock_gettime(CLOCK_MONOTONIC, &timer) != 0) {
        puts("Unable to get time for starting a timer, exiting");
        exit(-1);
    }
    return (((long long)timer.tv_nsec) >> 10) + ((long long)timer.tv_sec) * 1000000;
    */

    // This implementation with gettimeofday works and is tested,
    // but runs very slowly!
    struct timeval timer;
    if(gettimeofday(&timer, NULL) != 0) {
	puts("Unable to get time for starting a timer, exiting");
        exit(-1);
    }
    return (long long)timer.tv_usec + ((long long)timer.tv_sec) * 1000000;
}

long long BtiReadTimerMicro(long long startedTimer) {
    // This works and is tested, but is even slower than gettimeofday.
    // I believe that it is actually also less portable (not sure though)
    // However, I blieve that it should be more precise
    // (it is also nanoseconds, not microseconds)
    /*
    struct timespec timer;
    // CLOCK_REALTIME could also be used
    if(clock_gettime(CLOCK_MONOTONIC, &timer) != 0) {
        puts("Unable to get time for reading a timer, exiting");
        exit(-1);
    }
    return (((long long)timer.tv_nsec) >> 10) + ((long long)timer.tv_sec) * 1000000 - startedTimer;
    */

    // This implementation with gettimeofday works and is tested,
    // but runs very slowly!
    struct timeval timer;
    if(gettimeofday(&timer, NULL) != 0) {
	puts("Unable to get time for starting a timer, exiting");
        exit(-1);
    }
    return (long long)timer.tv_usec + ((long long)timer.tv_sec) * 1000000 - startedTimer;
}

void BtiStartPauseTimer() {
    btiPauseTimer = BtiStartTimer();
}

long long BtiReadPauseTimerMicro() {
    long long pauseTime = BtiReadTimerMicro(btiPauseTimer);

    if(pauseTime > btiMaxPauseTime) {
        btiMaxPauseTime = pauseTime;
    }
    btiTotalPauseTime += pauseTime;
    btiPauseCnt++;

    return pauseTime;
}

// A much quicker polling of whether the timer has fired than gettimeofday
int BtiIsQuickAsyncTimerFired() {
    // Alternative slow implementation which should work,
    // if using the precise clock_gettime implementation and it wasn't so slow!
    /*
    if(BtiReadTimerMicro(btiSlowIncrTimer) > 100000) {
	return 1;
    } else {
	return 0;
    }
    */
    // FIXME: Supposedly fast asynchronous implementation - possibly buggy
    return btiQuickAsyncTimerFired;
}

// A simple handler which just sets a flag, if the timer has fired
// FIXME: Supposedly fast asynchronous implementation - possibly buggy
static void BtiQuickAsyncTimerHandler(int sig, siginfo_t *si, void *uc) {
    btiQuickAsyncTimerFired = 1;
    signal(sig, SIG_IGN);
}

// Initialize the timer module
void BtiSetup() {
    struct sigaction sa;

    /* Establish handler for timer signal */

    sa.sa_flags = SA_SIGINFO;
    sa.sa_sigaction = BtiQuickAsyncTimerHandler;
    sigemptyset(&sa.sa_mask);
    if (sigaction(BTI_QUICK_ASYNC_TIMER_SIGNAL, &sa, NULL) == -1) {
        puts("Error while trying to setup signal handler for timer, exiting");
	exit(-1);
    }

    /* Create the timer */
    btiQuickAsyncTimerSev.sigev_notify = SIGEV_SIGNAL;
    btiQuickAsyncTimerSev.sigev_signo  = BTI_QUICK_ASYNC_TIMER_SIGNAL;
    btiQuickAsyncTimerSev.sigev_value.sival_ptr = &btiQuickAsyncTimerId;
    if (timer_create(CLOCK_REALTIME, &btiQuickAsyncTimerSev, &btiQuickAsyncTimerId) == -1) {
	puts("Error creating asynchronous timer, exiting\n");
    }
}

// Start quick asynchronous timer in micro seconds
void BtiStartQuickAsyncTimerMicro(long microSeconds) {
    struct sigaction sa;
    struct itimerspec its;

    // Alternative slow implementation which should work,
    // if using the precise clock_gettime implementation and it wasn't so slow!
    //btiSlowIncrTimer = BtiStartTimer();

    // FIXME: Supposedly fast asynchronous implementation - possibly buggy
    /* Establish handler for timer signal */
    sa.sa_flags = SA_SIGINFO;
    sa.sa_sigaction = BtiQuickAsyncTimerHandler;
    sigemptyset(&sa.sa_mask);
    if (sigaction(BTI_QUICK_ASYNC_TIMER_SIGNAL, &sa, NULL) == -1) {
        puts("Error while trying to setup signal handler for timer, exiting");
	exit(-1);
    }

    // Specify the timer
    its.it_value.tv_sec = 0;
    its.it_value.tv_nsec = microSeconds * 1000;
    its.it_interval.tv_sec = its.it_value.tv_sec;
    its.it_interval.tv_nsec = its.it_value.tv_nsec;

    // Start the timer
    if (timer_settime(btiQuickAsyncTimerId, 0, &its, NULL) == -1) {
	puts("Error starting asynchronous timer, exiting\n");
    }

    // Make sure that the timer fired flag is cleared
    btiQuickAsyncTimerFired = 0;
}
